@@ -220,6 +220,7 @@ static struct tegra_fb_data paz00_fb_data = {
 	.xres		= 1024,
 	.yres		= 600,
 	.bits_per_pixel	= 16,
+	.flags		= TEGRA_FB_FLIP_ON_PROBE,
 };
 
 static struct tegra_fb_data paz00_hdmi_fb_data = {
@@ -346,9 +347,11 @@ int __init paz00_panel_init(size_t fb_addr)
 	gpio_request_one(TEGRA_HDMI_HPD, GPIOF_IN, "hdmi_hpd");
 	gpio_export(TEGRA_LVDS_SHUTDOWN, 0);
 
-	paz00_disp1_resources[2].start = fb_addr;
-	/* 2.4 MB framebuffer should be enough */
-	paz00_disp1_resources[2].end =   fb_addr + 0x26B000 - 1;
+	if (fb_addr) {
+		paz00_disp1_resources[2].start = fb_addr;
+		/* 2.4 MB framebuffer should be enough */
+		paz00_disp1_resources[2].end =   fb_addr + 0x26B000 - 1;
+	}
 
 	err = platform_add_devices(paz00_gfx_devices,
 				   ARRAY_SIZE(paz00_gfx_devices));
