@@ -16,7 +16,7 @@
 
 
 #include <linux/delay.h>
-#include <linux/gpio.h>
+#include <linux/firmware.h>
 #include <linux/i2c.h>
 #include <linux/i2c/cyapa.h>
 #include <linux/input.h>
@@ -26,11 +26,9 @@
 #include <linux/miscdevice.h>
 #include <linux/module.h>
 #include <linux/mutex.h>
-#include <linux/semaphore.h>
 #include <linux/slab.h>
 #include <linux/spinlock.h>
 #include <linux/uaccess.h>
-#include <linux/workqueue.h>
 
 
 /* commands for read/write registers of Cypress trackpad */
@@ -211,15 +209,11 @@ struct cyapa {
 
 	struct i2c_client	*client;
 	struct input_dev	*input;
-	struct work_struct detect_work;
-	struct workqueue_struct *detect_wq;
 	int irq;
-	u8 adapter_func;
 	bool smbus;
 
 	/* read from query data region. */
 	char product_id[16];
-	u8 capability[14];
 	u8 fw_maj_ver;  /* firmware major version. */
 	u8 fw_min_ver;  /* firmware minor version. */
 	u8 hw_maj_ver;  /* hardware major version. */
@@ -337,6 +331,21 @@ static const struct cyapa_cmd_len cyapa_smbus_cmds[] = {
 	{CYAPA_SMBUS_BLK_HEAD, 16},
 };
 
+#define CYAPA_FW_NAME		"cyapa.bin"
+#define CYAPA_FW_BLOCK_SIZE	64
+#define CYAPA_FW_HDR_START	0x0780
+#define CYAPA_FW_HDR_BLOCK_COUNT  2
+#define CYAPA_FW_HDR_BLOCK_START  (CYAPA_FW_HDR_START / CYAPA_FW_BLOCK_SIZE)
+#define CYAPA_FW_HDR_SIZE	(CYAPA_FW_HDR_BLOCK_COUNT * \
+				 CYAPA_FW_BLOCK_SIZE)
+#define CYAPA_FW_DATA_START	0x0800
+#define CYAPA_FW_DATA_BLOCK_COUNT  480
+#define CYAPA_FW_DATA_BLOCK_START  (CYAPA_FW_DATA_START / CYAPA_FW_BLOCK_SIZE)
+#define CYAPA_FW_DATA_SIZE	(CYAPA_FW_DATA_BLOCK_COUNT * \
+				 CYAPA_FW_BLOCK_SIZE)
+#define CYAPA_FW_SIZE		(CYAPA_FW_HDR_SIZE + CYAPA_FW_DATA_SIZE)
+#define CYAPA_CMD_LEN		16
+
 static void cyapa_detect(struct cyapa *cyapa);
 
 #define BYTE_PER_LINE  8
@@ -473,7 +482,7 @@ static s32 cyapa_read_byte(struct cyapa *cyapa, u8 cmd_idx)
 	}
 	ret = i2c_smbus_read_byte_data(cyapa->client, cmd);
 	dev_dbg(dev, "read byte [0x%02x] = 0x%02x  ret: %d\n",
-		cmd, ret, ret);
+		cmd, (u8)ret, ret);
 
 	return ret;
 }
@@ -551,7 +560,7 @@ static int cyapa_get_state(struct cyapa *cyapa)
 	 * -ETIMEDOUT.  In this case, try again using the smbus equivalent
 	 * command.  This should return a BL_HEAD indicating CYAPA_STATE_OP.
 	 */
-	if (cyapa->smbus && ret == -ETIMEDOUT) {
+	if (cyapa->smbus && (ret == -ETIMEDOUT || ret == -ENXIO)) {
 		dev_dbg(dev, "smbus: probing with BL_STATUS command\n");
 		ret = cyapa_read_block(cyapa, CYAPA_CMD_BL_STATUS, status);
 	}
@@ -636,6 +645,7 @@ static int cyapa_bl_enter(struct cyapa *cyapa)
 	if (ret < 0)
 		return -EIO;
 
+	usleep_range(25000, 50000);
 	ret = cyapa_get_state(cyapa);
 	if (ret < 0)
 		return ret;
@@ -732,21 +742,18 @@ static int cyapa_set_power_mode(struct cyapa *cyapa, u8 power_mode)
 {
 	int ret;
 	u8 power;
-	int tries = 3;
 
 	if (cyapa->state != CYAPA_STATE_OP)
 		return 0;
 
-	power = cyapa_read_byte(cyapa, CYAPA_CMD_POWER_MODE);
+	ret = cyapa_read_byte(cyapa, CYAPA_CMD_POWER_MODE);
+	if (ret < 0)
+		return ret;
+
+	power = ret;
 	power &= ~OP_POWER_MODE_MASK;
 	power |= ((power_mode << OP_POWER_MODE_SHIFT) & OP_POWER_MODE_MASK);
-	do {
-		ret = cyapa_write_byte(cyapa, CYAPA_CMD_POWER_MODE, power);
-		/* sleep at least 10 ms. */
-		usleep_range(SET_POWER_MODE_DELAY, 2 * SET_POWER_MODE_DELAY);
-	} while ((ret != 0) && (tries-- > 0));
-
-	return ret;
+	return cyapa_write_byte(cyapa, CYAPA_CMD_POWER_MODE, power);
 }
 
 static int cyapa_get_query_data(struct cyapa *cyapa)
@@ -861,6 +868,212 @@ static int cyapa_check_is_operational(struct cyapa *cyapa)
 	return 0;
 }
 
+
+static u16 cyapa_csum(const u8 *buf, size_t count)
+{
+	int i;
+	u16 csum = 0;
+
+	for (i = 0; i < count; i++)
+		csum += buf[i];
+
+	return csum;
+}
+
+/*
+ * Write a |len| byte long buffer |buf| to the device, by chopping it up into a
+ * sequence of smaller |CYAPA_CMD_LEN|-length write commands.
+ *
+ * The data bytes for a write command are prepended with the 1-byte offset
+ * of the data relative to the start of |buf|.
+ */
+static int cyapa_write_buffer(struct cyapa *cyapa, const u8 *buf, size_t len)
+{
+	int ret;
+	size_t i;
+	unsigned char cmd[CYAPA_CMD_LEN + 1];
+	size_t cmd_len;
+
+	for (i = 0; i < len; i += CYAPA_CMD_LEN) {
+		const u8 *payload = &buf[i];
+		cmd_len = (len - i >= CYAPA_CMD_LEN) ? CYAPA_CMD_LEN : len - i;
+		cmd[0] = i;
+		memcpy(&cmd[1], payload, cmd_len);
+
+		ret = cyapa_i2c_reg_write_block(cyapa, 0, cmd_len + 1, cmd);
+		if (ret < 0)
+			return ret;
+	}
+	return 0;
+}
+
+
+/*
+ * A firmware block write command writes 64 bytes of data to a single flash
+ * page in the device.  The 78-byte block write command has the format:
+ *   <0xFF> <CMD> <Key> <Start> <Data> <Data-Checksum> <CMD Checksum>
+ *
+ *  <0xFF>  - every command starts with 0xFF
+ *  <CMD>   - the write command value is 0x39
+ *  <Key>   - write commands include an 8-byte key: { 00 01 02 03 04 05 06 07 }
+ *  <Block> - Memory Block number (address / 64) (16-bit, big-endian)
+ *  <Data>  - 64 bytes of firmware image data
+ *  <Data Checksum> - sum of 64 <Data> bytes, modulo 0xFF
+ *  <CMD Checksum> - sum of 77 bytes, from 0xFF to <Data Checksum>
+ *
+ * Each write command is split into 5 i2c write transactions of up to 16 bytes.
+ * Each transaction starts with an i2c register offset: (00, 10, 20, 30, 40).
+ */
+static int cyapa_write_fw_block(struct cyapa *cyapa, u16 block, const u8 *data)
+{
+	int ret;
+	u8 cmd[78];
+	u8 status[BL_STATUS_SIZE];
+	int tries = 3;
+
+	/* set write command and security key bytes. */
+	cmd[0] = 0xFF;
+	cmd[1] = 0x39;
+	cmd[2] = 0x00;
+	cmd[3] = 0x01;
+	cmd[4] = 0x02;
+	cmd[5] = 0x03;
+	cmd[6] = 0x04;
+	cmd[7] = 0x05;
+	cmd[8] = 0x06;
+	cmd[9] = 0x07;
+	cmd[10] = block >> 8;
+	cmd[11] = block;
+	memcpy(&cmd[12], data, CYAPA_FW_BLOCK_SIZE);
+	cmd[76] = cyapa_csum(data, CYAPA_FW_BLOCK_SIZE);
+	cmd[77] = cyapa_csum(cmd, sizeof(cmd) - 1);
+
+	ret = cyapa_write_buffer(cyapa, cmd, sizeof(cmd));
+	if (ret)
+		return ret;
+
+	/* wait for write to finish */
+	do {
+		usleep_range(10000, 20000);
+
+		/* check block write command result status. */
+		ret = cyapa_i2c_reg_read_block(cyapa, BL_HEAD_OFFSET,
+					       BL_STATUS_SIZE, status);
+		if (ret != BL_STATUS_SIZE)
+			return (ret < 0) ? ret : -EIO;
+		ret = (status[1] == 0x10 && status[2] == 0x20) ? 0 : -EIO;
+	} while (--tries && ret);
+
+	return ret;
+}
+
+
+/*
+ * Verify the integrity of a CYAPA firmware image file.
+ *
+ * The firmware image file is 30848 bytes, composed of 482 64-byte blocks.
+ *
+ * The first 2 blocks are the firmware header.
+ * The next 480 blocks are the firmware image.
+ *
+ * The first two bytes of the header hold the header checksum, computed by
+ * summing the other 126 bytes of the header.
+ * The last two bytes of the header hold the firmware image checksum, computed
+ * by summing the 30720 bytes of the image modulo 0xFFFF.
+ *
+ * Both checksums are stored little-endian.
+ */
+static int cyapa_check_fw(struct cyapa *cyapa, const struct firmware *fw)
+{
+	struct device *dev = &cyapa->client->dev;
+	u16 csum;
+	u16 csum_expected;
+
+	/* Firmware must match exact 30848 bytes = 482 64-byte blocks. */
+	if (fw->size != CYAPA_FW_SIZE) {
+		dev_err(dev, "invalid firmware size = %u, expected %u.\n",
+			fw->size, CYAPA_FW_SIZE);
+		return -EINVAL;
+	}
+
+	/* Verify header block */
+	csum_expected = (fw->data[0] << 8) | fw->data[1];
+	csum = cyapa_csum(&fw->data[2], CYAPA_FW_HDR_SIZE - 2);
+	if (csum != csum_expected) {
+		dev_err(dev, "invalid firmware header checksum = %04x, "
+			     "expected: %04x\n", csum, csum_expected);
+		return -EINVAL;
+	}
+
+	/* Verify firmware image */
+	csum_expected = (fw->data[CYAPA_FW_HDR_SIZE - 2] << 8) |
+			 fw->data[CYAPA_FW_HDR_SIZE - 1];
+	csum = cyapa_csum(&fw->data[CYAPA_FW_HDR_SIZE], CYAPA_FW_DATA_SIZE);
+	if (csum != csum_expected) {
+		dev_err(dev, "invalid firmware checksum = %04x, "
+			     "expected: %04x\n", csum, csum_expected);
+		return -EINVAL;
+	}
+	return 0;
+}
+
+static int cyapa_firmware(struct cyapa *cyapa)
+{
+	struct device *dev = &cyapa->client->dev;
+	int ret;
+	const struct firmware *fw;
+	const char *fw_name = CYAPA_FW_NAME;
+	int i;
+
+	ret = request_firmware(&fw, CYAPA_FW_NAME, dev);
+	if (ret) {
+		dev_err(dev, "Could not load firmware from %s, %d\n",
+			  fw_name, ret);
+		return ret;
+	}
+
+	ret = cyapa_check_fw(cyapa, fw);
+	if (ret) {
+		dev_err(dev, "Invalid CYAPA firmware image: %s\n", fw_name);
+		goto done;
+	}
+
+	ret = cyapa_bl_enter(cyapa);
+	if (ret)
+		goto err_detect;
+
+	ret = cyapa_bl_activate(cyapa);
+	if (ret)
+		goto err_detect;
+
+	/* First write data, starting at byte 128  of fw->data */
+	for (i = 0; i < CYAPA_FW_DATA_BLOCK_COUNT; i++) {
+		size_t block = CYAPA_FW_DATA_BLOCK_START + i;
+		size_t addr = (i + CYAPA_FW_HDR_BLOCK_COUNT) *
+				CYAPA_FW_BLOCK_SIZE;
+		const u8 *data = &fw->data[addr];
+		ret = cyapa_write_fw_block(cyapa, block, data);
+		if (ret)
+			goto err_detect;
+	}
+
+	/* Then write checksum */
+	for (i = 0; i < CYAPA_FW_HDR_BLOCK_COUNT; i++) {
+		size_t block = CYAPA_FW_HDR_BLOCK_START + i;
+		size_t addr = i * CYAPA_FW_BLOCK_SIZE;
+		const u8 *data = &fw->data[addr];
+		ret = cyapa_write_fw_block(cyapa, block, data);
+		if (ret)
+			goto err_detect;
+	}
+
+err_detect:
+	cyapa_detect(cyapa);
+
+done:
+	release_firmware(fw);
+	return ret;
+}
 /*
  **************************************************************
  * misc cyapa device for trackpad firmware update,
@@ -1064,6 +1277,7 @@ static int cyapa_send_bl_cmd(struct cyapa *cyapa, enum cyapa_bl_cmd cmd)
 
 	case CYAPA_CMD_IDLE_TO_APP:
 		cyapa_detect(cyapa);
+		ret = 0;
 		break;
 
 	default:
@@ -1266,16 +1480,35 @@ static ssize_t cyapa_show_protocol_version(struct device *dev,
 	return sprintf(buf, "%d\n", cyapa->gen);
 }
 
+static ssize_t cyapa_update_fw_store(struct device *dev,
+				     struct device_attribute *attr,
+				     const char *buf, size_t count)
+{
+	struct cyapa *cyapa = dev_get_drvdata(dev);
+	int ret;
+
+	ret = cyapa_firmware(cyapa);
+	if (ret)
+		dev_err(dev, "firmware update failed, %d\n", ret);
+	else
+		dev_dbg(dev, "firmware update succeeded\n");
+
+	return ret ? ret : count;
+}
+
+
 static DEVICE_ATTR(firmware_version, S_IRUGO, cyapa_show_fm_ver, NULL);
 static DEVICE_ATTR(hardware_version, S_IRUGO, cyapa_show_hw_ver, NULL);
 static DEVICE_ATTR(product_id, S_IRUGO, cyapa_show_product_id, NULL);
 static DEVICE_ATTR(protocol_version, S_IRUGO, cyapa_show_protocol_version, NULL);
+static DEVICE_ATTR(update_fw, S_IWUSR, NULL, cyapa_update_fw_store);
 
 static struct attribute *cyapa_sysfs_entries[] = {
 	&dev_attr_firmware_version.attr,
 	&dev_attr_hardware_version.attr,
 	&dev_attr_product_id.attr,
 	&dev_attr_protocol_version.attr,
+	&dev_attr_update_fw.attr,
 	NULL,
 };
 
@@ -1476,19 +1709,9 @@ static void cyapa_detect(struct cyapa *cyapa)
 		if (ret)
 			dev_err(dev, "create input_dev instance failed, %d\n",
 				ret);
-	} else {
-		ret = cyapa_set_power_mode(cyapa, PWR_MODE_FULL_ACTIVE);
-		if (ret)
-			dev_warn(dev, "resume active power failed, %d\n", ret);
 	}
 }
 
-static void cyapa_detect_work(struct work_struct *work)
-{
-	struct cyapa *cyapa = container_of(work, struct cyapa, detect_work);
-	cyapa_detect(cyapa);
-}
-
 static int __devinit cyapa_probe(struct i2c_client *client,
 				 const struct i2c_device_id *dev_id)
 {
@@ -1513,9 +1736,8 @@ static int __devinit cyapa_probe(struct i2c_client *client,
 	cyapa->client = client;
 	i2c_set_clientdata(client, cyapa);
 
-	cyapa->adapter_func = adapter_func;
 	/* i2c isn't supported, set smbus */
-	if (cyapa->adapter_func == CYAPA_ADAPTER_FUNC_SMBUS)
+	if (adapter_func == CYAPA_ADAPTER_FUNC_SMBUS)
 		cyapa->smbus = true;
 	cyapa->state = CYAPA_STATE_NO_DEVICE;
 
@@ -1546,30 +1768,10 @@ static int __devinit cyapa_probe(struct i2c_client *client,
 	if (sysfs_create_group(&client->dev.kobj, &cyapa_sysfs_group))
 		dev_warn(dev, "error creating sysfs entries.\n");
 
-	/*
-	 * At boot it can take up to 2 seconds for firmware to complete sensor
-	 * calibration. Probe in a workqueue so as not to block system boot.
-	 */
-	cyapa->detect_wq = create_singlethread_workqueue("cyapa_detect_wq");
-	if (!cyapa->detect_wq) {
-		ret = -ENOMEM;
-		dev_err(dev, "create detect workqueue failed\n");
-		goto err_irq_free;
-	}
-
-	INIT_WORK(&cyapa->detect_work, cyapa_detect_work);
-	ret = queue_work(cyapa->detect_wq, &cyapa->detect_work);
-	if (ret < 0) {
-		dev_err(dev, "device detect failed, %d\n", ret);
-		goto err_wq_free;
-	}
+	cyapa_detect(cyapa);
 
 	return 0;
 
-err_wq_free:
-	destroy_workqueue(cyapa->detect_wq);
-err_irq_free:
-	free_irq(cyapa->irq, cyapa);
 err_mem_free:
 	kfree(cyapa);
 	global_cyapa = NULL;
@@ -1588,8 +1790,6 @@ static int __devexit cyapa_remove(struct i2c_client *client)
 	if (cyapa->input)
 		input_unregister_device(cyapa->input);
 
-	if (cyapa->detect_wq)
-		destroy_workqueue(cyapa->detect_wq);
 	kfree(cyapa);
 	global_cyapa = NULL;
 
@@ -1602,9 +1802,6 @@ static int cyapa_suspend(struct device *dev)
 	int ret;
 	struct cyapa *cyapa = dev_get_drvdata(dev);
 
-	/* Wait for detection to complete before allowing suspend. */
-	flush_workqueue(cyapa->detect_wq);
-
 	/* set trackpad device to light sleep mode. Just ignore any errors */
 	ret = cyapa_set_power_mode(cyapa, PWR_MODE_LIGHT_SLEEP);
 	if (ret < 0)
@@ -1612,6 +1809,7 @@ static int cyapa_suspend(struct device *dev)
 
 	if (device_may_wakeup(dev))
 		enable_irq_wake(cyapa->irq);
+	disable_irq(cyapa->irq);
 
 	return 0;
 }
@@ -1621,15 +1819,15 @@ static int cyapa_resume(struct device *dev)
 	int ret;
 	struct cyapa *cyapa = dev_get_drvdata(dev);
 
+	enable_irq(cyapa->irq);
 	if (device_may_wakeup(dev))
 		disable_irq_wake(cyapa->irq);
 
-	PREPARE_WORK(&cyapa->detect_work, cyapa_detect_work);
-	ret = queue_work(cyapa->detect_wq, &cyapa->detect_work);
-	if (ret < 0) {
-		dev_err(dev, "queue detect work failed, %d\n", ret);
-		return ret;
-	}
+	cyapa_detect(cyapa);
+
+	ret = cyapa_set_power_mode(cyapa, PWR_MODE_FULL_ACTIVE);
+	if (ret)
+		dev_warn(dev, "resume active power failed, %d\n", ret);
 
 	return 0;
 }
@@ -1643,11 +1841,19 @@ static const struct i2c_device_id cyapa_id_table[] = {
 };
 MODULE_DEVICE_TABLE(i2c, cyapa_id_table);
 
+static const struct of_device_id cyapa_of_match[] = {
+	{ .compatible = "cy,i2c-apa", },
+	{ .compatible = "cy,apa", },
+	{ },
+};
+MODULE_DEVICE_TABLE(of, cyapa_of_match);
+
 static struct i2c_driver cyapa_driver = {
 	.driver = {
 		.name = CYAPA_I2C_NAME,
 		.owner = THIS_MODULE,
 		.pm = &cyapa_pm_ops,
+		.of_match_table = cyapa_of_match,
 	},
 
 	.probe = cyapa_probe,
